from typing import Optional
import copy

from .attribute import AttributeType
from .component import Component
from .rule import Rules, RuleType


class Panel:

    def __init__(self,
                 component_1: Component,
                 component_2: Optional[Component] = None):

        self.component_1 = component_1
        if component_2:
            self.component_2 = component_2
            self.components = (self.component_1, self.component_2)
        else:
            self.components = (self.component_1, )

    def __str__(self):
        s = f"{self.component_1.config}\n{self.component_1.uniformity}\n\n"
        for entity in self.component_1.entities:
            s += f"{entity!r}\n"
        if hasattr(self, "component_2"):
            s += f"\n\t----------- \\\\ {self.component_1.component_type.name} > {self.component_1.layout_type.name} // --- " + \
                 f"// {self.component_2.component_type.name} > {self.component_2.layout_type.name} \\\\ -----------\n\n"
            s += f"{self.component_2.config}\n{self.component_2.uniformity}\n"
            for entity in self.component_2.entities:
                s += f"{entity!r}\n"
        return s


def prune(base: Panel, rules: Rules) -> Optional[Panel]:
    """
    Modify the bounds of attributes based on the rules that will be applied
    to them to ensure those rules can be properly expressed with the given
    range of values; if this cannot be done by tightening bounds, returns None.
    If rules cannot be applied, the panel is reset such that `self.constraints`
    does not differ from its state before the method was called.

    Note that the bounds in base.components[i].constraints.number are the maximal
    and minimal number of objects that may be placed in this layout.  In particular,

        base.components[i].constraints.number == \
            base.components[i].config.position.value.shape[0]
     
    """
    pruned = copy.deepcopy(base)
    for component, component_rules in zip(pruned.components, rules):
        del component.entities[:]
        for rule in component_rules.all:
            if rule.attr in AttributeType and \
                    rule.attr is not AttributeType.ANGLE and \
                    rule.attr is not AttributeType.UNIFORMITY:
                if rule.attr is AttributeType.NUMBER or rule.attr is AttributeType.POSITION or \
                    rule.attr is AttributeType.CONFIGURATION:
                    bounds = getattr(component.constraints,
                                     AttributeType.NUMBER.name.lower())
                else:
                    bounds = getattr(component.constraints,
                                     rule.attr.name.lower())
                initial_min, initial_max = bounds.min, bounds.max
                if rule.name is RuleType.PROGRESSION:
                    if rule.attr is AttributeType.POSITION:
                        # can increment/decrement layout slot indices twice without repetition:
                        #   bounds.max >= bounds.min + 2 * abs(rule.value)
                        bounds.max = bounds.max - 2 * abs(rule.value)
                    elif bounds:
                        # bounds.max >= bounds.min + 2 * rule.value
                        if rule.value > 0:
                            bounds.max = bounds.max - 2 * rule.value
                        else:
                            bounds.min = bounds.min - 2 * rule.value
                elif rule.name is RuleType.ARITHMETIC:
                    if rule.attr is AttributeType.POSITION:
                        # sets must be incomplete (union) and overlapping (difference)
                        if rule.value <= 0:
                            bounds.min = bounds.max // 2
                        bounds.max = bounds.max - 1
                    elif rule.attr is AttributeType.NUMBER or rule.attr is AttributeType.SIZE:
                        # bounds.max >= 2 * bounds.min + 1
                        if rule.value > 0:
                            bounds.max = bounds.max - bounds.min - 1
                        else:
                            bounds.min = 2 * bounds.min + 1
                    elif rule.attr is AttributeType.COLOR:
                        # bounds.max >= bounds.min + 1
                        #   && bounds.max >= 2 * bounds.min
                        if bounds.max - bounds.min < 1:
                            return None
                        elif rule.value > 0:
                            bounds.max = bounds.max - bounds.min
                        elif rule.value < 0:
                            bounds.min = 2 * bounds.min
                elif rule.name is RuleType.DISTRIBUTE_THREE:
                    if rule.attr is AttributeType.POSITION:
                        # n choose k entity arrangements in layout with n slots;
                        # `value_of_setting(bounds.min)` >= 1 and we restrict
                        # n >= 3 and k < n so n choose k >= 3 (see Pascal's triangle).
                        n_slots = component.config.number.value_of_setting(
                            bounds.max)
                        if n_slots < 3:
                            return None
                        else:
                            bounds.max = bounds.max - 1
                    elif bounds:
                        # u - l + 1 settings in [u, l]: require >= 3
                        if bounds.max - bounds.min + 1 < 3:
                            return None
                if bounds.max < bounds.min:  # reset bounds if pruned
                    bounds.min, bounds.max = initial_min, initial_max
                    return None
            else:
                return None
    return pruned



from typing import List
from enum import Enum, auto
import copy, os, math

import numpy as np
from scipy.special import comb
from PIL import Image

from .component import make_component, ComponentType, LayoutType, AttributeHistory
from .panel import Panel, prune
from . import attribute
from .attribute import AttributeType, PositionType, PlanarPosition
from .rule import Ruleset, RuleType, apply_rule


class StructureType(Enum):
    SINGLETON = auto()
    LEFT_RIGHT = auto()
    UP_DOWN = auto()
    OUT_IN = auto()


class MatrixType(Enum):
    ONE_SHAPE = auto()
    FOUR_SHAPE = auto()
    FIVE_SHAPE = auto()
    NINE_SHAPE = auto()
    TWO_SHAPE_VERTICAL_SEP = auto()
    TWO_SHAPE_HORIZONTAL_SEP = auto()
    SHAPE_IN_SHAPE = auto()
    FOUR_SHAPE_IN_SHAPE = auto()


class Matrix:

    attribute_bounds = {
        MatrixType.ONE_SHAPE: {
            (ComponentType.NONE, LayoutType.CENTER): {}
        },
        MatrixType.FOUR_SHAPE: {
            (ComponentType.NONE, LayoutType.GRID_FOUR): {}
        },
        MatrixType.FIVE_SHAPE: {
            (ComponentType.NONE, LayoutType.GRID_FIVE): {}
        },
        MatrixType.NINE_SHAPE: {
            (ComponentType.NONE, LayoutType.GRID_NINE): {}
        },
        MatrixType.TWO_SHAPE_VERTICAL_SEP: {
            (ComponentType.LEFT, LayoutType.CENTER): {},
            (ComponentType.RIGHT, LayoutType.CENTER): {}
        },
        MatrixType.TWO_SHAPE_HORIZONTAL_SEP: {
            (ComponentType.UP, LayoutType.CENTER): {},
            (ComponentType.DOWN, LayoutType.CENTER): {}
        },
        MatrixType.SHAPE_IN_SHAPE: {
            (ComponentType.OUT, LayoutType.CENTER): {},
            (ComponentType.IN, LayoutType.CENTER): {}
        },
        MatrixType.FOUR_SHAPE_IN_SHAPE: {
            (ComponentType.OUT, LayoutType.CENTER): {},
            (ComponentType.IN, LayoutType.GRID_FOUR): {}
        }
    }

    @staticmethod
    def oblique_angle_rotations(allowed: bool):
        if allowed:
            if attribute.ANGLE_VALUES != (-135, -90, -45, 0, 45, 90, 135, 180):
                attribute.ANGLE_VALUES = (-135, -90, -45, 0, 45, 90, 135, 180)
                attribute.ANGLE_MIN, attribute.ANGLE_MAX = 0, len(
                    attribute.ANGLE_VALUES) - 1
        else:
            if attribute.ANGLE_VALUES != (-90, 0, 90, 180):
                attribute.ANGLE_VALUES = (-90, 0, 90, 180)
                attribute.ANGLE_MIN, attribute.ANGLE_MAX = 0, len(
                    attribute.ANGLE_VALUES) - 1

    @classmethod
    def make(cls,
             matrix_type: MatrixType,
             ruleset: Ruleset = None,
             n_alternatives: int = 0):
        return getattr(cls, matrix_type.name.lower())(
            cls.get_attribute_bounds(matrix_type), ruleset, n_alternatives)

    @staticmethod
    def validate_bounds(src, dst, attribute: AttributeType, min_setting,
                        max_setting):
        if attribute is AttributeType.NUMBER:
            min_key, max_key = "number_min", "number_max"
        elif attribute is AttributeType.SHAPE:
            min_key, max_key = "shape_min", "shape_max"
        elif attribute is AttributeType.SIZE:
            min_key, max_key = "size_min", "size_max"
        elif attribute is AttributeType.COLOR:
            min_key, max_key = "color_min", "color_max"
        elif attribute is AttributeType.ANGLE:
            min_key, max_key = "angle_min", "angle_max"
        elif attribute is AttributeType.UNIFORMITY:
            min_key, max_key = "uniformity_min", "uniformity_max"
        if min_key not in src:
            dst[min_key] = min_setting
        else:
            dst[min_key] = max(min(src[min_key], max_setting), min_setting)
        if max_key not in src:
            dst[max_key] = max_setting
        else:
            dst[max_key] = max(min(src[max_key], max_setting), min_setting)
        if dst[max_key] < dst[min_key]:
            dst[max_key] = dst[min_key]

    @classmethod
    def get_attribute_bounds(cls, matrix_type: MatrixType):
        validated = {}
        for (component_type,
             layout_type), src in cls.attribute_bounds[matrix_type].items():
            assert (component_type in ComponentType
                    and layout_type in LayoutType)
            dst = {}
            if layout_type is LayoutType.CENTER:
                number_max = attribute.NUM_MIN
            elif layout_type is LayoutType.GRID_FOUR:
                number_max = attribute.NUM_MIN + 3
            elif layout_type is LayoutType.GRID_FIVE:
                number_max = attribute.NUM_MIN + 4
            else:  # layout_type is LayoutType.GRID_NINE
                number_max = attribute.NUM_MIN + 8
            cls.validate_bounds(src, dst, AttributeType.NUMBER,
                                attribute.NUM_MIN, number_max)
            cls.validate_bounds(src, dst, AttributeType.SHAPE,
                                attribute.SHAPE_MIN + 1, attribute.SHAPE_MAX)
            if component_type is ComponentType.OUT:
                size_min, color_max = attribute.SIZE_MIN + 3, attribute.COLOR_MIN
            elif component_type is ComponentType.IN and layout_type is LayoutType.GRID_FOUR:
                size_min = attribute.SIZE_MIN + 2
            else:
                size_min, color_max = attribute.SIZE_MIN, attribute.COLOR_MAX
            cls.validate_bounds(src, dst, AttributeType.SIZE, size_min,
                                attribute.SIZE_MAX)
            cls.validate_bounds(src, dst, AttributeType.COLOR,
                                attribute.COLOR_MIN, color_max)
            cls.validate_bounds(src, dst, AttributeType.ANGLE,
                                attribute.ANGLE_MIN, attribute.ANGLE_MAX)
            cls.validate_bounds(src, dst, AttributeType.UNIFORMITY,
                                attribute.UNI_MIN, attribute.UNI_MAX)
            validated[(component_type, layout_type)] = dst
        cls.attribute_bounds[matrix_type] = validated
        return validated

    @classmethod
    def one_shape(cls, bounds, ruleset, n_alternatives):
        return cls(
            StructureType.SINGLETON,
            Panel(
                make_component(
                    ComponentType.NONE, LayoutType.CENTER, PositionType.PLANAR,
                    [PlanarPosition(x_c=0.5, y_c=0.5, max_w=1, max_h=1)],
                    **(bounds[(ComponentType.NONE, LayoutType.CENTER)]))),
            ruleset, n_alternatives)

    @classmethod
    def four_shape(cls, bounds, ruleset, n_alternatives):
        return cls(
            StructureType.SINGLETON,
            Panel(
                make_component(
                    ComponentType.NONE, LayoutType.GRID_FOUR,
                    PositionType.PLANAR, [
                        PlanarPosition(
                            x_c=0.25, y_c=0.25, max_w=0.5, max_h=0.5),
                        PlanarPosition(
                            x_c=0.25, y_c=0.75, max_w=0.5, max_h=0.5),
                        PlanarPosition(
                            x_c=0.75, y_c=0.25, max_w=0.5, max_h=0.5),
                        PlanarPosition(
                            x_c=0.75, y_c=0.75, max_w=0.5, max_h=0.5)
                    ],
                    **(bounds[(ComponentType.NONE, LayoutType.GRID_FOUR)]))),
            ruleset, n_alternatives)

    @classmethod
    def five_shape(cls, bounds, ruleset, n_alternatives):
        return cls(
            StructureType.SINGLETON,
            Panel(
                make_component(
                    ComponentType.NONE, LayoutType.GRID_FIVE,
                    PositionType.PLANAR, [
                        PlanarPosition(
                            x_c=0.2, y_c=0.2, max_w=0.33, max_h=0.33),
                        PlanarPosition(
                            x_c=0.2, y_c=0.8, max_w=0.33, max_h=0.33),
                        PlanarPosition(
                            x_c=0.5, y_c=0.5, max_w=0.33, max_h=0.33),
                        PlanarPosition(
                            x_c=0.8, y_c=0.2, max_w=0.33, max_h=0.33),
                        PlanarPosition(
                            x_c=0.8, y_c=0.8, max_w=0.33, max_h=0.33)
                    ],
                    **(bounds[(ComponentType.NONE, LayoutType.GRID_FIVE)]))),
            ruleset, n_alternatives)

    @classmethod
    def nine_shape(cls, bounds, ruleset, n_alternatives):
        return cls(
            StructureType.SINGLETON,
            Panel(
                make_component(
                    ComponentType.NONE, LayoutType.GRID_NINE,
                    PositionType.PLANAR, [
                        PlanarPosition(
                            x_c=0.16, y_c=0.16, max_w=0.33, max_h=0.33),
                        PlanarPosition(
                            x_c=0.16, y_c=0.5, max_w=0.33, max_h=0.33),
                        PlanarPosition(
                            x_c=0.16, y_c=0.83, max_w=0.33, max_h=0.33),
                        PlanarPosition(
                            x_c=0.5, y_c=0.16, max_w=0.33, max_h=0.33),
                        PlanarPosition(
                            x_c=0.5, y_c=0.5, max_w=0.33, max_h=0.33),
                        PlanarPosition(
                            x_c=0.5, y_c=0.83, max_w=0.33, max_h=0.33),
                        PlanarPosition(
                            x_c=0.83, y_c=0.16, max_w=0.33, max_h=0.33),
                        PlanarPosition(
                            x_c=0.83, y_c=0.5, max_w=0.33, max_h=0.33),
                        PlanarPosition(
                            x_c=0.83, y_c=0.83, max_w=0.33, max_h=0.33)
                    ],
                    **(bounds[(ComponentType.NONE, LayoutType.GRID_NINE)]))),
            ruleset, n_alternatives)

    @classmethod
    def two_shape_vertical_sep(cls, bounds, ruleset, n_alternatives):
        return cls(
            StructureType.LEFT_RIGHT,
            Panel(
                make_component(
                    ComponentType.LEFT, LayoutType.CENTER, PositionType.PLANAR,
                    [PlanarPosition(x_c=0.5, y_c=0.25, max_w=0.5, max_h=0.5)],
                    **(bounds[(ComponentType.LEFT, LayoutType.CENTER)])),
                make_component(
                    ComponentType.RIGHT, LayoutType.CENTER,
                    PositionType.PLANAR,
                    [PlanarPosition(x_c=0.5, y_c=0.75, max_w=0.5, max_h=0.5)],
                    **(bounds[(ComponentType.RIGHT, LayoutType.CENTER)]))),
            ruleset, n_alternatives)

    @classmethod
    def two_shape_horizontal_sep(cls, bounds, ruleset, n_alternatives):
        return cls(
            StructureType.UP_DOWN,
            Panel(
                make_component(
                    ComponentType.UP, LayoutType.CENTER, PositionType.PLANAR,
                    [PlanarPosition(x_c=0.25, y_c=0.5, max_w=0.5, max_h=0.5)],
                    **(bounds[(ComponentType.UP, LayoutType.CENTER)])),
                make_component(
                    ComponentType.DOWN, LayoutType.CENTER, PositionType.PLANAR,
                    [PlanarPosition(x_c=0.75, y_c=0.5, max_w=0.5, max_h=0.5)],
                    **(bounds[(ComponentType.DOWN, LayoutType.CENTER)]))),
            ruleset, n_alternatives)

    @classmethod
    def shape_in_shape(cls, bounds, ruleset, n_alternatives):
        return cls(
            StructureType.OUT_IN,
            Panel(
                make_component(
                    ComponentType.OUT, LayoutType.CENTER, PositionType.PLANAR,
                    [PlanarPosition(x_c=0.5, y_c=0.5, max_w=1, max_h=1)],
                    **(bounds[(ComponentType.OUT, LayoutType.CENTER)])),
                make_component(
                    ComponentType.IN, LayoutType.CENTER, PositionType.PLANAR,
                    [PlanarPosition(x_c=0.5, y_c=0.5, max_w=0.33, max_h=0.33)],
                    **(bounds[(ComponentType.IN, LayoutType.CENTER)]))),
            ruleset, n_alternatives)

    @classmethod
    def four_shape_in_shape(cls, bounds, ruleset, n_alternatives):
        return cls(
            StructureType.OUT_IN,
            Panel(
                make_component(
                    ComponentType.OUT, LayoutType.CENTER, PositionType.PLANAR,
                    [PlanarPosition(x_c=0.5, y_c=0.5, max_w=1, max_h=1)],
                    **(bounds[(ComponentType.OUT, LayoutType.CENTER)])),
                make_component(
                    ComponentType.IN, LayoutType.GRID_FOUR,
                    PositionType.PLANAR, [
                        PlanarPosition(
                            x_c=0.42, y_c=0.42, max_w=0.15, max_h=0.15),
                        PlanarPosition(
                            x_c=0.42, y_c=0.58, max_w=0.15, max_h=0.15),
                        PlanarPosition(
                            x_c=0.58, y_c=0.42, max_w=0.15, max_h=0.15),
                        PlanarPosition(
                            x_c=0.58, y_c=0.58, max_w=0.15, max_h=0.15)
                    ], **(bounds[(ComponentType.IN, LayoutType.GRID_FOUR)]))),
            ruleset, n_alternatives)

    def __init__(self, structure_type, base, ruleset, n_alternatives):
        self.structure_type = structure_type
        self.initial_constraints = tuple(
            copy.deepcopy(comp.constraints) for comp in base.components)
        self.make_ground_truth(base, ruleset)
        self.make_alternatives(n_alternatives)

    def make_ground_truth(self, base, ruleset):
        if ruleset is None:
            ruleset = Ruleset()
        while True:
            self.rules = ruleset.sample(
                tuple((c.component_type, c.layout_type)
                      for c in base.components))
            pruned = prune(base, self.rules)
            self.start = copy.deepcopy(
                pruned) if pruned is not None else pruned
            if self.start is not None:
                break

        panels = []
        for _ in range(3):
            panels.extend(self.make_row())

        self.context, self.answer = panels[:-1], panels[-1]

    def make_row(self):
        col_0 = copy.deepcopy(self.start)
        for comp in col_0.components:
            comp.sample(sample_position=True,
                        sample_number=True,
                        carryover=False)
        comp_rows = []
        for comp_rules, comp_0 in zip(self.rules, col_0.components):
            comp_1 = apply_rule(comp_rules.configuration, comp_0)
            comp_2 = apply_rule(comp_rules.configuration, comp_1)
            for rule in comp_rules:
                comp_1 = apply_rule(rule, comp_0, comp_1)
            for rule in comp_rules:
                comp_2 = apply_rule(rule, comp_1, comp_2)
            comp_rows.append((comp_0, comp_1, comp_2))
        if len(comp_rows) == 1:
            comp_rows.append((None, None, None))
        return [Panel(comp_1, comp_2) for (comp_1, comp_2) in zip(*comp_rows)]

    def make_alternatives(self, n_alternatives):
        self.alternatives = []
        if n_alternatives > 0:
            self.modifications = self.count_modifiable()
            self.uniques = [
                AttributeHistory(cst) for cst in self.initial_constraints
            ]
            for _ in range(n_alternatives):
                if len(self.modifications) == 0:
                    break
                c, attr = self.sample_modification()
                alternative = copy.deepcopy(self.answer)
                alternative.components[c].sample_unique(
                    attr, self.uniques[c], self.initial_constraints[c])
                self.alternatives.append(alternative)

    def count_modifiable(self) -> List[list]:
        samples = []
        for c, (component, component_rules, initial_constraints) in enumerate(
                zip(self.answer.components, self.rules,
                    self.initial_constraints)):
            n_samples, max_entities = 0, component.config.position.values.shape[
                0]
            for n_entities in range(
                    component.config.number.value_of_setting(
                        initial_constraints.number.min),
                    component.config.number.value_of_setting(
                        initial_constraints.number.max) + 1):
                if n_entities != component.config.number.value:
                    n_samples += comb(max_entities, n_entities)
            if n_samples > 0:
                samples.append([c, AttributeType.NUMBER, n_samples])
            if component_rules.configuration.attr is not AttributeType.NUMBER:
                position_samples = comb(max_entities,
                                        component.config.number.value) - 1
                if position_samples > 0:
                    samples.append(
                        [c, AttributeType.POSITION, position_samples])
            for rule in component_rules:
                if rule.name is not RuleType.CONSTANT or \
                    component.uniformity.value or \
                    component_rules.configuration.name is RuleType.CONSTANT or \
                    (component_rules.configuration.attr is AttributeType.POSITION and
                     component_rules.configuration.name is not RuleType.ARITHMETIC):
                    bounds = getattr(initial_constraints,
                                     rule.attr.name.lower())
                    n_samples = bounds.max - bounds.min  # excludes one setting
                    if not component.uniformity.value:
                        n_samples //= len(component.entities)
                    if n_samples > 0:
                        samples.append([c, rule.attr, n_samples])
        return samples

    def sample_modification(self):
        i = np.random.choice(len(self.modifications))
        component_idx, attr_name, available_samples = self.modifications[i]
        if available_samples - 1 == 0:
            del self.modifications[i]
        else:
            self.modifications[i][2] = available_samples - 1
        return component_idx, attr_name

    def render(self, panel, background_color, panel_size, line_thickness,
               shape_border_thickness):
        entities = []
        for component in panel.components:
            entities.extend(component.entities)
        panel_img = np.ones(
            (panel_size, panel_size), np.uint8) * background_color
        for entity in entities:
            entity_img = entity.render(background_color, panel_size,
                                       shape_border_thickness)
            panel_img[entity_img != background_color] = 0
            entity_img[entity_img == background_color] = 0
            panel_img += entity_img
        neg_offset, pos_offset = math.floor(line_thickness / 2), math.ceil(
            line_thickness / 2)
        band_center = int(0.5 * panel_size)
        if self.structure_type is StructureType.LEFT_RIGHT:
            panel_img[:,
                      band_center - neg_offset:band_center + pos_offset] = 0.
        elif self.structure_type is StructureType.UP_DOWN:
            panel_img[band_center - neg_offset:band_center +
                      pos_offset, :] = 0.
        return panel_img

    def __str__(self):
        s = f"\nstructure: {self.structure_type.name}:\n"
        for i, panel in enumerate(self.context):
            if (i % 3) == 0:
                s += "\n\t+++++++++++ ROW +++++++++++\n\n"
            elif i != 0:
                s += "\n\t*********** PANEL ***********\n\n"
            s += str(panel)
        s += "\n\t*********** SOLUTION ***********\n\n"
        s += str(self.answer)
        for alternative in self.alternatives:
            s += "\n\t*********** ALTERNATIVE ***********\n\n"
            s += str(alternative)
        return s

    def generate_matrix(self, last_panel, background_color, panel_size,
                        line_thickness, shape_border_thickness):
        img_grid = np.ones(
            (panel_size * 3, panel_size * 3), np.uint8) * background_color
        for pos, panel in enumerate([
                self.render(panel, background_color, panel_size,
                            line_thickness, shape_border_thickness)
                for panel in self.context
        ] + [
                self.render(last_panel, background_color, panel_size,
                            line_thickness, shape_border_thickness)
        ]):
            i, j = divmod(pos, 3)
            img_grid[i * panel_size:(i + 1) * panel_size,
                     j * panel_size:(j + 1) * panel_size] = panel
        neg_offset, pos_offset = math.floor(line_thickness / 2), math.ceil(
            line_thickness / 2)
        for x in [0.33, 0.67]:
            band_center = int(x * panel_size * 3)
            img_grid[band_center - neg_offset:band_center + pos_offset, :] = 0
        for y in [0.33, 0.67]:
            band_center = int(y * panel_size * 3)
            img_grid[:, band_center - neg_offset:band_center + pos_offset] = 0
        return Image.fromarray(img_grid)

    def save(self,
             path,
             puzzle_name,
             background_color=255,
             image_size=480,
             line_thickness=3,
             shape_border_thickness=2):
        img = self.generate_matrix(self.answer, background_color,
                                   image_size // 3, line_thickness,
                                   shape_border_thickness)
        img.save(os.path.join(path, puzzle_name + "_answer.png"))
        for i, alternative in enumerate(self.alternatives):
            img = self.generate_matrix(alternative, background_color,
                                       image_size // 3, line_thickness,
                                       shape_border_thickness)
            img.save(os.path.join(path, puzzle_name + f"_alternative_{i}.png"))



from typing import Union
from dataclasses import dataclass, field

import cv2
import numpy as np

from .attribute import (Shape, Size, Color, Angle, Shapes, PlanarPosition,
                        AngularPosition)


@dataclass
class Point:
    x: Union[int, float]
    y: Union[int, float]

    def __iter__(self):
        for attr in ["y", "x"]:
            yield getattr(self, attr)


def rotate(img, angle, background_color, panel_size, center=None):
    if center is None:
        center = Point(x=(panel_size // 2), y=(panel_size // 2))
    return cv2.warpAffine(img,
                          cv2.getRotationMatrix2D(tuple(center), angle, 1),
                          (panel_size, panel_size),
                          flags=cv2.INTER_LINEAR,
                          borderValue=background_color)


@dataclass(init=False)
class Entity:
    name: str
    bbox: Union[PlanarPosition, AngularPosition] = field(repr=False)
    shape: Shape
    size: Size
    color: Color
    angle: Angle

    def __init__(self, name, bbox, constraints):
        self.name = name
        self.bbox = bbox
        self.shape = Shape(constraints)
        self.size = Size(constraints)
        self.color = Color(constraints)
        self.angle = Angle(constraints)

    def sample(self, constraints):
        self.shape.sample(constraints)
        self.size.sample(constraints)
        self.color.sample(constraints)
        self.angle.sample(constraints)

    def render(self, background_color, panel_size, shape_border_thickness):
        img = np.ones((panel_size, panel_size), np.uint8) * background_color
        center = Point(y=int(self.bbox.y_c * panel_size),
                       x=int(self.bbox.x_c * panel_size))
        unit = min(self.bbox.max_w, self.bbox.max_h) * panel_size // 2
        # minus because of the way we show the image, see render_panel's return
        if self.shape.value is Shapes.TRIANGLE:
            dl = int(unit * self.size.value)
            pts = np.array([[
                center.y, center.x - dl
            ], [
                center.y + int(dl / 2.0 * np.sqrt(3)), center.x + int(dl / 2.0)
            ], [
                center.y - int(dl / 2.0 * np.sqrt(3)), center.x + int(dl / 2.0)
            ]], np.int32).reshape((-1, 1, 2))
            if self.color.value != background_color:
                cv2.fillConvexPoly(img, pts, self.color.value)
            cv2.polylines(img, [pts], True, 0, shape_border_thickness)
        elif self.shape.value is Shapes.SQUARE:
            dl = int(unit / 2 * np.sqrt(2) * self.size.value)
            pt1 = (center.y - dl, center.x - dl)
            pt2 = (center.y + dl, center.x + dl)
            if self.color.value != background_color:
                cv2.rectangle(img, pt1, pt2, self.color.value, -1)
            cv2.rectangle(img, pt1, pt2, 0, shape_border_thickness)
        elif self.shape.value is Shapes.PENTAGON:
            dl = int(unit * self.size.value)
            pts = np.array([[center.y, center.x - dl],
                            [
                                center.y - int(dl * np.cos(np.pi / 10)),
                                center.x - int(dl * np.sin(np.pi / 10))
                            ],
                            [
                                center.y - int(dl * np.sin(np.pi / 5)),
                                center.x + int(dl * np.cos(np.pi / 5))
                            ],
                            [
                                center.y + int(dl * np.sin(np.pi / 5)),
                                center.x + int(dl * np.cos(np.pi / 5))
                            ],
                            [
                                center.y + int(dl * np.cos(np.pi / 10)),
                                center.x - int(dl * np.sin(np.pi / 10))
                            ]], np.int32).reshape((-1, 1, 2))
            if self.color.value != background_color:
                cv2.fillConvexPoly(img, pts, self.color.value)
            cv2.polylines(img, [pts], True, 0, shape_border_thickness)
        elif self.shape.value is Shapes.HEXAGON:
            dl = int(unit * self.size.value)
            pts = np.array([[
                center.y, center.x - dl
            ], [
                center.y - int(dl / 2.0 * np.sqrt(3)), center.x - int(dl / 2.0)
            ], [
                center.y - int(dl / 2.0 * np.sqrt(3)), center.x + int(dl / 2.0)
            ], [
                center.y, center.x + dl
            ], [
                center.y + int(dl / 2.0 * np.sqrt(3)), center.x + int(dl / 2.0)
            ], [
                center.y + int(dl / 2.0 * np.sqrt(3)), center.x - int(dl / 2.0)
            ]], np.int32).reshape((-1, 1, 2))
            if self.color.value != background_color:
                cv2.fillConvexPoly(img, pts, self.color.value)
            cv2.polylines(img, [pts], True, 0, shape_border_thickness)
        elif self.shape.value is Shapes.CIRCLE:
            radius = int(unit * self.size.value)
            if self.color.value != background_color:
                cv2.circle(img, tuple(center), radius, self.color.value, -1)
            cv2.circle(img, tuple(center), radius, 0, shape_border_thickness)
        elif self.shape.value is Shapes.NONE:
            pass
        if isinstance(self.bbox, AngularPosition):
            return rotate(img,
                          self.bbox.omega,
                          background_color,
                          panel_size,
                          center=Point(x=(self.bbox.x_r * panel_size),
                                       y=(self.bbox.y_r * panel_size)))
        elif isinstance(self.bbox, PlanarPosition):
            return rotate(img,
                          self.angle.value,
                          background_color,
                          panel_size,
                          center=center)
        else:
            raise ValueError("unknown position type: not angular or planar")



from typing import Union
from dataclasses import dataclass
from enum import Enum, auto

import numpy as np

NUM_VALUES = (1, 2, 3, 4, 5, 6, 7, 8, 9)
NUM_MIN = 0
NUM_MAX = len(NUM_VALUES) - 1

UNI_VALUES = (False, False, False, True)
UNI_MIN = 0
UNI_MAX = len(UNI_VALUES) - 1


class Shapes(Enum):
    NONE = auto()
    TRIANGLE = auto()
    SQUARE = auto()
    PENTAGON = auto()
    HEXAGON = auto()
    CIRCLE = auto()


SHAPE_VALUES = tuple(Shapes)
SHAPE_MIN = 0
SHAPE_MAX = len(SHAPE_VALUES) - 1

SIZE_VALUES = (0.4, 0.5, 0.6, 0.7, 0.8, 0.9)
SIZE_MIN = 0
SIZE_MAX = len(SIZE_VALUES) - 1

COLOR_VALUES = (255, 224, 196, 168, 140, 112, 84, 56, 28, 0)
COLOR_MIN = 0
COLOR_MAX = len(COLOR_VALUES) - 1

ANGLE_VALUES = (-135, -90, -45, 0, 45, 90, 135, 180)
ANGLE_MIN = 0
ANGLE_MAX = len(ANGLE_VALUES) - 1


class AttributeType(Enum):
    NUMBER = auto()
    SHAPE = auto()
    SIZE = auto()
    COLOR = auto()
    ANGLE = auto()
    UNIFORMITY = auto()
    POSITION = auto()
    CONFIGURATION = auto()


class Attribute:

    def __init__(self, name: AttributeType, values):
        self.name = name
        self.values = np.array(values, dtype="object")

    @property
    def value(self):
        return self.values[self.setting]

    def value_of_setting(self, setting):
        return self.values[setting]

    def reset(self):
        self.setting = self.initial_setting

    def __repr__(self):
        return f"{self.name.name.title()}(setting={self.setting}, value={self.value}))"

    def __str__(self):
        return f"{self.name.name.title()}(setting={self.setting}, value={self.value}))"


class Sampleable(Attribute):

    def __init__(self, name: AttributeType, values, constraints):
        super(Sampleable, self).__init__(name, values)
        self.sample(constraints)
        self.initial_setting = self.setting

    def sample(self, constraints):
        constraint = getattr(constraints, self.name.name.lower())
        self.setting = np.random.choice(
            range(constraint.min, constraint.max + 1))


class UniqueSampleable(Sampleable):

    def sample_unique(self, constraints, history, record=True, inplace=False):
        constraint = getattr(constraints, self.name.name.lower())
        previous_settings = getattr(history, self.name.name.lower())
        all_settings = range(constraint.min, constraint.max + 1)
        new_setting = np.random.choice(
            list(
                set(all_settings) - set(previous_settings) -
                set((self.setting, ))))
        if record:
            if self.setting not in previous_settings:
                previous_settings.append(self.setting)
            previous_settings.append(new_setting)
        if inplace:
            self.setting = new_setting
        return new_setting


class Number(UniqueSampleable):

    def __init__(self, constraints):
        super(Number, self).__init__(name=AttributeType.NUMBER,
                                     values=NUM_VALUES,
                                     constraints=constraints)


class PositionType(Enum):
    PLANAR = auto()
    ANGULAR = auto()


@dataclass
class PlanarPosition:
    x_c: Union[int, float]
    y_c: Union[int, float]
    max_w: Union[int, float]
    max_h: Union[int, float]


@dataclass
class AngularPosition:
    x_c: Union[int, float]
    y_c: Union[int, float]
    max_w: Union[int, float]
    max_h: Union[int, float]
    x_r: Union[int, float]
    y_r: Union[int, float]
    omega: Union[int, float]


class Position(Attribute):

    def __init__(self, constraints, n_entities: int):
        self.position_type = constraints.position_type
        super(Position, self).__init__(name=AttributeType.POSITION,
                                       values=constraints.positions)
        self.sample(n_entities)

    def sample(self, size):
        assert size <= self.values.shape[0]
        self.setting = np.random.choice(self.values.shape[0],
                                        size=size,
                                        replace=False)

    def record_setting(self, history, setting, positions_set=None):
        position_history = getattr(history, self.name.name.lower())
        position_history[len(setting)].available -= 1
        position_history[len(setting)].sampled.append(
            set(setting) if positions_set is None else positions_set)

    def sample_unique(self, size, history, record=True, inplace=False):
        position_history = getattr(history, self.name.name.lower())
        new_setting = np.random.choice(self.values.shape[0],
                                       size=size,
                                       replace=False)
        while True:
            unique, new_positions_set = True, set(new_setting)
            current_positions_set = set(self.setting)
            if new_positions_set in position_history[size].sampled or \
               new_positions_set == current_positions_set:
                unique = False
            if unique:
                if record:
                    if current_positions_set not in position_history[len(
                            self.setting)].sampled:
                        self.record_setting(
                            history,
                            self.setting,
                            positions_set=current_positions_set)
                    self.record_setting(history,
                                        new_setting,
                                        positions_set=new_positions_set)
                if inplace:
                    self.setting = new_setting
                return new_setting
            else:
                new_setting = np.random.choice(self.values.shape[0],
                                               size=size,
                                               replace=False)


class Configuration:

    def __init__(self, constraints):
        self.number = Number(constraints)
        self.position = Position(constraints, self.number.value)

    def __repr__(self):
        return f"Configuration(number={self.number!r}, position={self.position!r})"

    def __str__(self):
        return f"Configuration(number={self.number!r}, position={self.position!r})"

    def sample(self, constraints):
        self.number.sample(constraints)
        self.position.sample(self.number.value)

    def sample_unique(self, constraints, history, inplace=False):
        while True:
            number_setting = self.number.sample_unique(constraints,
                                                       history,
                                                       record=False)
            number_value = self.number.value_of_setting(number_setting)
            if history.position[number_value].available == 0:
                continue
            position_setting = self.position.sample_unique(number_value,
                                                           history,
                                                           record=False)
            position_setting_set = set(position_setting)
            if position_setting_set not in history.position[
                    number_value].sampled:
                self.position.record_setting(history, self.position.setting)
                self.position.record_setting(
                    history,
                    position_setting,
                    positions_set=position_setting_set)
                if self.number.setting not in history.number:
                    history.number.append(self.number.setting)
                history.number.append(number_setting)
                if inplace:
                    self.number.setting = number_setting
                    self.position.setting = position_setting
                return self.number.setting, self.position.setting

    def reset(self):
        self.number.reset()
        self.position.reset()


class Shape(UniqueSampleable):

    def __init__(self, constraints):
        super(Shape, self).__init__(name=AttributeType.SHAPE,
                                    values=SHAPE_VALUES,
                                    constraints=constraints)


class Size(UniqueSampleable):

    def __init__(self, constraints):
        super(Size, self).__init__(name=AttributeType.SIZE,
                                   values=SIZE_VALUES,
                                   constraints=constraints)


class Color(UniqueSampleable):

    def __init__(self, constraints):
        super(Color, self).__init__(name=AttributeType.COLOR,
                                    values=COLOR_VALUES,
                                    constraints=constraints)


class Angle(UniqueSampleable):

    def __init__(self, constraints):
        super(Angle, self).__init__(name=AttributeType.ANGLE,
                                    values=ANGLE_VALUES,
                                    constraints=constraints)


class Uniformity(Sampleable):

    def __init__(self, constraints):
        super(Uniformity, self).__init__(name=AttributeType.UNIFORMITY,
                                         values=UNI_VALUES,
                                         constraints=constraints)



from typing import Iterable, List, Optional, Tuple
from enum import Enum, auto
from dataclasses import dataclass, field
import copy

import numpy as np

from .attribute import AttributeType, COLOR_MIN, COLOR_MAX
from .component import Component, ComponentType, LayoutType


class RuleType(Enum):
    CONSTANT = auto()
    PROGRESSION = auto()
    ARITHMETIC = auto()
    DISTRIBUTE_THREE = auto()


@dataclass
class Rule:
    name: RuleType
    attr: AttributeType
    params: List[int]
    value: Optional[int] = field(default=None, init=False)

    def __post_init__(self):
        if self.params is not None:
            self.value = np.random.choice(self.params)

    def increment(self):
        pass


def apply_rule(rule: Rule,
               prev_comp: Component,
               next_comp: Optional[Component] = None) -> Component:
    if next_comp is None:
        next_comp = prev_comp
    if rule.attr is AttributeType.ANGLE or \
            rule.attr is AttributeType.UNIFORMITY or \
            rule.attr not in AttributeType:
        raise ValueError("unsupported attribute")
    elif rule.name is RuleType.CONSTANT:
        next_comp = copy.deepcopy(next_comp)
        if rule.attr is AttributeType.CONFIGURATION:
            next_comp.sample(carryover=False)
        else:
            if rule.prev_is_col_0 and not prev_comp.uniformity.value:
                prev_comp.make_uniform(rule.attr)
            next_comp.set_uniform(rule.attr, prev_comp.setting_of(rule.attr))
    elif rule.name is RuleType.PROGRESSION:
        next_comp = copy.deepcopy(next_comp)
        if rule.attr is AttributeType.NUMBER:
            next_comp.config.number.setting += rule.value
            next_comp.sample(sample_position=True, carryover=False)
        elif rule.attr is AttributeType.POSITION:
            next_comp.config.position.setting = (
                next_comp.config.position.setting +
                rule.value) % next_comp.config.position.values.shape[0]
            next_comp.sample(carryover=False)
        else:
            if rule.prev_is_col_0 and not prev_comp.uniformity.value:
                prev_comp.make_uniform(rule.attr)
            next_comp.set_uniform(rule.attr,
                                  prev_comp.setting_of(rule.attr) + rule.value)
    elif rule.name is RuleType.ARITHMETIC:
        if rule.attr is AttributeType.SHAPE:
            raise ValueError("unsupported attribute")
        next_comp = copy.deepcopy(next_comp)
        if rule.attr is AttributeType.NUMBER:
            if rule.col_0_setting is None:  # second column
                rule.col_0_setting = prev_comp.config.number.setting
                rule.set_constraints_col_1(prev_comp, next_comp)
                next_comp.config.number.sample(next_comp.constraints)
            else:  # third column
                next_comp.config.number.setting = rule.col_2_setting(
                    prev_comp.config.number.setting)
            next_comp.sample(sample_position=True, carryover=False)
        elif rule.attr is AttributeType.POSITION:
            if rule.col_0_setting is None:  # second column
                rule.col_0_setting = prev_comp.config.position.setting
                rule.set_constraints_col_1(prev_comp, next_comp)
            else:  # third column
                col_2_setting = rule.col_2_setting(
                    prev_comp.config.position.setting)
                next_comp.config.number.setting = len(col_2_setting) - 1
                next_comp.config.position.setting = np.array(col_2_setting)
            next_comp.sample(carryover=False)
        elif rule.attr is AttributeType.SIZE:
            if rule.col_0_setting is None:  # second column
                rule.col_0_setting = prev_comp.setting_of(rule.attr)
                if not prev_comp.uniformity.value:
                    prev_comp.make_uniform(rule.attr)
                rule.set_constraints_col_1(prev_comp, next_comp)
                next_comp.attr(rule.attr).sample(next_comp.constraints)
                next_comp.make_uniform(rule.attr)
            else:  # third column
                next_comp.set_uniform(
                    rule.attr,
                    rule.col_2_setting(prev_comp.setting_of(rule.attr)))
        elif rule.attr is AttributeType.COLOR:
            rule.color_count += 1
            if rule.col_0_setting is None:
                prev_setting = prev_comp.setting_of(rule.attr)
                reset_previous = False
                if rule.color_count == 3 and rule.color_white_alarm and \
                    ((rule.value > 0 and prev_setting == COLOR_MAX) or
                        (rule.value < 0 and prev_setting == COLOR_MIN)):
                    prev_comp.attr(rule.attr).sample_unique(
                        prev_comp.constraints, prev_comp.history, inplace=True)
                    prev_setting = prev_comp.setting_of(rule.attr)
                    reset_previous = True
                rule.col_0_setting = prev_setting
                if reset_previous or not prev_comp.uniformity.value:
                    prev_comp.make_uniform(rule.attr)
                rule.set_constraints_col_1(prev_comp, next_comp)
                next_comp.attr(rule.attr).sample(next_comp.constraints)
                next_setting = next_comp.setting_of(rule.attr)
                if rule.color_count == 1:
                    rule.color_white_alarm = (next_setting == 0)
                elif rule.color_count == 3 and rule.color_white_alarm and \
                     next_setting == 0:
                    next_comp.attr(rule.attr).sample_unique(
                        next_comp.constraints, next_comp.history, inplace=True)
                next_comp.make_uniform(rule.attr)
            else:  # third column
                next_comp.set_uniform(
                    rule.attr,
                    rule.col_2_setting(prev_comp.setting_of(rule.attr)))
    elif rule.name is RuleType.DISTRIBUTE_THREE:
        if rule.attr is AttributeType.NUMBER:
            if rule.count == 0:  # first row
                rule.create_settings(
                    np.insert(
                        np.random.choice(
                            list(
                                range(prev_comp.constraints.number.min,
                                      prev_comp.config.number.setting)) +
                            list(
                                range(prev_comp.config.number.setting + 1,
                                      prev_comp.constraints.number.max + 1)),
                            size=2,
                            replace=False), 0,
                        prev_comp.config.number.setting))
                next_comp = copy.deepcopy(next_comp)
                next_comp.config.number.setting = rule.settings[0][1]
            else:
                row, col = divmod(rule.count, 2)
                if col == 0:
                    prev_comp.config.number.setting = rule.settings[row][0]
                    prev_comp.sample(sample_position=True)
                    next_comp = copy.deepcopy(prev_comp)
                    next_comp.config.number.setting = rule.settings[row][1]
                else:
                    next_comp = copy.deepcopy(next_comp)
                    next_comp.config.number.setting = rule.settings[row][2]
            next_comp.sample(sample_position=True)
        elif rule.attr is AttributeType.POSITION:
            if rule.count == 0:
                rule.create_settings(
                    np.array([
                        prev_comp.config.position.setting,
                        prev_comp.config.position.sample_unique(
                            prev_comp.config.number.value, prev_comp.history),
                        prev_comp.config.position.sample_unique(
                            prev_comp.config.number.value, prev_comp.history)
                    ]))
                next_comp = copy.deepcopy(next_comp)
                next_comp.config.position.setting = rule.settings[0][1]
            else:
                row, col = divmod(rule.count, 2)
                if col == 0:
                    prev_comp.config.number.setting = rule.settings[row][
                        0].shape[0] - 1
                    prev_comp.config.position.setting = rule.settings[row][0]
                    prev_comp.sample()
                    next_comp = copy.deepcopy(prev_comp)
                    next_comp.config.position.setting = rule.settings[row][1]
                else:
                    next_comp = copy.deepcopy(next_comp)
                    next_comp.config.position.setting = rule.settings[row][2]
            next_comp.sample()
        else:

            def attr_of(x):
                return getattr(x, rule.attr.name.lower())

            next_comp = copy.deepcopy(next_comp)
            if rule.count == 0:
                rule.create_settings(
                    np.random.choice(a=range(
                        attr_of(prev_comp.constraints).min,
                        attr_of(prev_comp.constraints).max + 1),
                                     size=3,
                                     replace=False))
                prev_comp.set_uniform(rule.attr, rule.settings[0][0])
                next_comp.set_uniform(rule.attr, rule.settings[0][1])
            else:
                row, col = divmod(rule.count, 2)
                if col == 0:
                    prev_comp.set_uniform(rule.attr, rule.settings[row][0])
                    next_comp.set_uniform(rule.attr, rule.settings[row][1])
                else:
                    next_comp.set_uniform(rule.attr, rule.settings[row][2])
    rule.increment()
    return next_comp


@dataclass
class Constant(Rule):

    def __post_init__(self):
        super(Constant, self).__post_init__()
        self.prev_is_col_0 = True

    def increment(self):
        self.prev_is_col_0 = not self.prev_is_col_0


@dataclass
class Progression(Rule):

    def __post_init__(self):
        super(Progression, self).__post_init__()
        self.prev_is_col_0 = True

    def increment(self):
        self.prev_is_col_0 = not self.prev_is_col_0


@dataclass
class Arithmetic(Rule):

    def __post_init__(self):
        super(Arithmetic, self).__post_init__()
        self.col_0_setting = None
        self.color_count = 0
        self.color_white_alarm = False

    def pop(self):
        col_0_setting = self.col_0_setting
        self.col_0_setting = None
        return col_0_setting

    def set_constraints_col_1(self, prev_comp, next_comp):
        if self.attr is AttributeType.POSITION:
            while True:
                next_comp.config.sample(next_comp.constraints)
                col_1_setting = next_comp.config.position.setting
                if self.value > 0:
                    if not (set(self.col_0_setting) >= set(col_1_setting)):
                        return
                else:
                    if not (set(self.col_0_setting) <= set(col_1_setting)):
                        return
        else:

            def attr_of(x):
                return getattr(x, self.attr.name.lower())

            next_constraint = attr_of(next_comp.constraints)
            prev_constraint = attr_of(prev_comp.constraints)
            if self.value > 0:
                next_constraint.max = prev_constraint.max - (
                    self.col_0_setting - prev_constraint.min)
            else:
                offset = 0 if self.attr is AttributeType.COLOR else 1
                min_prepruning = (next_constraint.min - offset) // 2
                next_constraint.min = min_prepruning
                if self.attr is AttributeType.COLOR:
                    next_constraint.max = self.col_0_setting
                else:
                    next_constraint.max = self.col_0_setting - min_prepruning - 1

    def col_2_setting(self, col_1_setting):
        col_0_setting = self.pop()
        if self.attr is AttributeType.POSITION:
            if self.value > 0:
                return list(set(col_0_setting) | set(col_1_setting))
            else:
                return list(set(col_0_setting) - set(col_1_setting))
        else:
            offset = 0 if self.attr is AttributeType.COLOR else 1
            if self.value > 0:
                return col_0_setting + (col_1_setting + offset)
            else:
                return col_0_setting - (col_1_setting + offset)


@dataclass
class DistributeThree(Rule):

    def __post_init__(self):
        super(DistributeThree, self).__post_init__()
        self.settings = []
        self.count = 0

    def create_settings(self, three_settings):
        self.settings.append(three_settings[[0, 1, 2]])
        if np.random.uniform() >= 0.5:
            self.settings.append(three_settings[[1, 2, 0]])
            self.settings.append(three_settings[[2, 0, 1]])
        else:
            self.settings.append(three_settings[[2, 0, 1]])
            self.settings.append(three_settings[[1, 2, 0]])

    def increment(self):
        self.count = (self.count + 1) % 6


@dataclass
class ComponentRules:
    configuration: Rule
    shape: Rule
    size: Rule
    color: Rule
    component_type: ComponentType
    layout_type: LayoutType

    def __post_init__(self):
        self.all = (self.configuration, self.shape, self.size, self.color)
        self.secondary = self.all[1:]

    def __iter__(self):
        for rule in self.secondary:
            yield rule


@dataclass(frozen=True)
class Rules:
    components_rules: Tuple[ComponentRules]

    def __iter__(self):
        for component_rules in self.components_rules:
            yield component_rules

    def __getitem__(self, i):
        return self.components_rules[i]

    def __len__(self):
        return len(self.components_rules)

    def __str__(self):
        s = "\n"
        for c, comp_rules in enumerate(self):
            for rule in comp_rules.all:
                s += f"{rule!r}\n"
            if c == 0 and len(self) > 1:
                s += f"\n\t----------- \\\\ {self[0].component_type.name} > {self[0].layout_type.name} // --- " + \
                     f"// {self[1].component_type.name} > {self[1].layout_type.name} \\\\ -----------\n\n"
        return s


@dataclass
class Ruleset:

    position_rules: Iterable[RuleType] = None
    number_rules: Iterable[RuleType] = None
    shape_rules: Iterable[RuleType] = None
    size_rules: Iterable[RuleType] = None
    color_rules: Iterable[RuleType] = None

    def __post_init__(self):
        for attr_rules in ("position_rules", "number_rules", "shape_rules",
                           "size_rules", "color_rules"):
            rules = getattr(self, attr_rules)
            if rules is None:
                validated_rules = []
            else:
                validated_rules = tuple(rule for rule in rules
                                        if rule in RuleType)
            if len(validated_rules) > 0:
                setattr(self, attr_rules, validated_rules)
            else:
                exclusions = {RuleType.ARITHMETIC
                              } if attr_rules == "shape_rules" else set()
                setattr(self, attr_rules, tuple(set(RuleType) - exclusions))

    @staticmethod
    def rule(name, attr):
        if name is RuleType.CONSTANT:
            return Constant(name, attr, params=None)
        elif name is RuleType.PROGRESSION:
            return Progression(name, attr, params=[-2, -1, 1, 2])
        elif name is RuleType.ARITHMETIC:
            return Arithmetic(name, attr, params=[-1, 1])
        else:  # name is RuleType.DISTRIBUTE_THREE
            return DistributeThree(name, attr, params=None)

    def sample(self, component_and_layout_types):
        assert (len(component_and_layout_types) == 1
                or len(component_and_layout_types) == 2)
        components_rules = []
        for component_type, layout_type in component_and_layout_types:
            name, attr = (np.random.choice(self.position_rules),
                          AttributeType.POSITION) if np.random.choice(
                              (True, False)) else (np.random.choice(
                                  self.number_rules), AttributeType.NUMBER)
            components_rules.append(
                ComponentRules(
                    self.rule(
                        name, AttributeType.CONFIGURATION
                        if name is RuleType.CONSTANT else attr),
                    self.rule(np.random.choice(self.shape_rules),
                              AttributeType.SHAPE),
                    self.rule(np.random.choice(self.size_rules),
                              AttributeType.SIZE),
                    self.rule(np.random.choice(self.color_rules),
                              AttributeType.COLOR), component_type,
                    layout_type))
        return Rules(tuple(components_rules))



from __future__ import annotations
import copy
from dataclasses import dataclass, field
from typing import List, Union, Dict, Set
from enum import Enum, auto

import numpy as np
from scipy.special import comb

from .attribute import (AttributeType, Uniformity, Configuration, PositionType,
                        AngularPosition, PlanarPosition)
from .entity import Entity


class ComponentType(Enum):
    NONE = auto()
    LEFT = auto()
    RIGHT = auto()
    UP = auto()
    DOWN = auto()
    OUT = auto()
    IN = auto()


class LayoutType(Enum):
    CENTER = auto()
    GRID_FOUR = auto()
    GRID_FIVE = auto()
    GRID_NINE = auto()


@dataclass
class PositionHistory:
    available: int
    sampled: List[Set[int]]


@dataclass(init=False)
class AttributeHistory:
    number: List[int]
    position: Dict[int, PositionHistory]
    shape: List[int]
    size: List[int]
    color: List[int]
    angle: List[int]

    def __init__(self, constraints):
        self.number = []
        self.position = {}
        self.shape = []
        self.size = []
        self.color = []
        self.angle = []
        for i in range(constraints.number.min + 1, constraints.number.max + 2):
            self.position[i] = PositionHistory(available=comb(
                constraints.positions.shape[0], i),
                                               sampled=[])


@dataclass
class Bounds:
    min: int
    max: int


@dataclass
class Constraints:
    number: Bounds
    shape: Bounds
    size: Bounds
    color: Bounds
    angle: Bounds
    uniformity: Bounds
    position_type: PositionType
    positions: Union[List[AngularPosition], List[PlanarPosition]]

    def __post_init__(self):
        self.positions = np.array(self.positions)


@dataclass
class Component:
    component_type: ComponentType
    layout_type: LayoutType
    constraints: Constraints
    config: Configuration = field(init=False)
    uniformity: Uniformity = field(init=False)
    history: AttributeHistory = field(init=False)
    entities: List[Entity] = field(init=False)
    initial_constraints: Constraints = field(init=False)

    def __post_init__(self):
        self.config = Configuration(self.constraints)
        self.uniformity = Uniformity(self.constraints)
        self.history = AttributeHistory(self.constraints)
        self.entities = []
        self.initial_constraints = copy.deepcopy(self.constraints)

    def setting_of(self, attr):
        return self.attr(attr).setting

    def attr(self, attr):
        return getattr(self.entity, attr.name.lower())

    @property
    def entity(self):
        return self.entities[0]

    def make_uniform(self, attr):
        setting = self.setting_of(attr)
        for entity in self.entities[1:]:
            entity_attr = getattr(entity, attr.name.lower())
            entity_attr.setting = setting

    def set_uniform(self, attr, setting):
        for entity in self.entities:
            entity_attr = getattr(entity, attr.name.lower())
            entity_attr.setting = setting

    def set_position(self):
        for entity, bbox in zip(self.entities, self.config.position.value):
            entity.bbox = bbox

    def sample(self,
               sample_position=False,
               sample_number=False,
               carryover=True,
               uniform=None):
        if sample_position or sample_number:
            if sample_number:
                self.config.number.sample(self.constraints)
            self.config.position.sample(self.config.number.value)
        if uniform is None:
            uniform = self.uniformity.value
        if uniform:
            if len(self.entities) > 0 and carryover:
                self.entities = [self.entities[0]]
            else:
                self.entities = [
                    Entity(name=str(0),
                           bbox=self.config.position.value[0],
                           constraints=self.constraints)
                ]
            for i, bbox in enumerate(self.config.position.value[1:]):
                entity = copy.deepcopy(self.entities[0])
                entity.name = str(i)
                entity.bbox = bbox
                self.entities.append(entity)
        else:
            self.entities = [
                Entity(name=str(i), bbox=bbox, constraints=self.constraints)
                for i, bbox in enumerate(self.config.position.value)
            ]

    def sample_unique(self, attr, history, initial_constraints):
        if attr is AttributeType.NUMBER:
            self.config.sample_unique(initial_constraints,
                                      history,
                                      inplace=True)
            self.sample(carryover=False)
        elif attr is AttributeType.POSITION:
            self.config.position.sample_unique(self.config.number.value,
                                               history,
                                               inplace=True)
            self.set_position()
        elif attr is AttributeType.ANGLE or \
                attr is AttributeType.UNIFORMITY:
            raise ValueError(f"unsupported operation attribute: {attr!s}")
        elif attr in AttributeType:
            if self.uniformity.value:
                self.attr(attr).sample_unique(initial_constraints,
                                              history,
                                              inplace=True)
                self.make_uniform(attr)
            else:
                for entity in self.entities:
                    entity_attr = getattr(entity, attr.name.lower())
                    entity_attr.sample_unique(initial_constraints,
                                              history,
                                              inplace=True)
        else:
            raise ValueError("unsupported operation")


def make_component(component_type, layout_type, position_type, positions, *,
                   number_min, number_max, shape_min, shape_max, size_min,
                   size_max, color_min, color_max, angle_min, angle_max,
                   uniformity_min, uniformity_max):
    return Component(component_type=component_type,
                     layout_type=layout_type,
                     constraints=Constraints(number=Bounds(min=number_min,
                                                           max=number_max),
                                             shape=Bounds(min=shape_min,
                                                          max=shape_max),
                                             size=Bounds(min=size_min,
                                                         max=size_max),
                                             color=Bounds(min=color_min,
                                                          max=color_max),
                                             angle=Bounds(min=angle_min,
                                                          max=angle_max),
                                             uniformity=Bounds(
                                                 min=uniformity_min,
                                                 max=uniformity_max),
                                             position_type=position_type,
                                             positions=positions))
